﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [StorableClass]
  public class Vertex : Item, IVertex {
    [Storable]
    private string id;
    public string Id {
      get { return id; }
      set { id = value; }
    }
    [Storable]
    private string label;
    public string Label { get { return label; } set { label = value; } }
    [Storable]
    private double weight;
    public double Weight { get { return weight; } set { weight = value; } }
    [Storable]
    private List<IArc> inArcs;
    public List<IArc> InArcs { get { return inArcs; } protected set { inArcs = value; } }
    [Storable]
    private List<IArc> outArcs;
    public List<IArc> OutArcs { get { return outArcs; } protected set { outArcs = value; } }

    [StorableConstructor]
    public Vertex(bool deserializing) : base(deserializing) { }

    public Vertex() {
      id = Guid.NewGuid().ToString();
    }

    protected Vertex(Vertex original, Cloner cloner)
      : base(original, cloner) {
      Id = Guid.NewGuid().ToString();
      Label = original.Label;
      InArcs = new List<IArc>(original.InArcs);
      OutArcs = new List<IArc>(original.OutArcs);
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new Vertex(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    protected void AfterDeserialization() {
      if (Id == null) {
        Id = Guid.NewGuid().ToString();
      }
    }

    // this method can be used to add arcs towards targets that are not visible in the graph 
    // (they do not appear in the nodes Dictionary). It is the programmers responsibility to 
    // enforce the correct and desired behavior.
    public void AddForwardArc(IVertex target, double weight = 0.0, object data = null) {
      var e = new Arc { Source = this, Target = target, Data = data, Weight = weight };
      if (OutArcs == null) OutArcs = new List<IArc> { e };
      else OutArcs.Add(e);
    }
    public void AddForwardArc(IArc arc) {
      if (arc.Source == null) { arc.Source = this; }
      if (arc.Source != this) { throw new Exception("AddForwardArc: Source should be equal to this."); }
      if (OutArcs == null) { OutArcs = new List<IArc> { arc }; } else { OutArcs.Add(arc); }
    }
    public void AddReverseArc(IArc arc) {
      if (arc.Target == null) { arc.Target = this; }
      if (arc.Target != this) { throw new Exception("AddReverseArc: Target should be equal to this."); };
      if (InArcs == null) { InArcs = new List<IArc> { arc }; } else { InArcs.Add(arc); }
    }
    public void AddReverseArc(IVertex source, double weight = 0.0, object data = null) {
      var e = new Arc { Source = source, Target = this, Data = data, Weight = weight };
      if (InArcs == null) InArcs = new List<IArc> { e };
      else InArcs.Add(e);
    }

    public int InDegree { get { return InArcs == null ? 0 : InArcs.Count; } }
    public int OutDegree { get { return OutArcs == null ? 0 : OutArcs.Count; } }
    public int Degree { get { return InDegree + OutDegree; } }
  }


  [StorableClass]
  public class Vertex<T> : Vertex, IVertex<T> where T : class,IItem {
    [Storable]
    private T content;
    public T Content {
      get { return content; }
      set {
        if (value == null)
          throw new ArgumentException("Node content cannot be null.");
        content = value;
      }
    }

    public Vertex() { }

    [StorableConstructor]
    protected Vertex(bool deserializing) : base(deserializing) { }

    protected Vertex(Vertex<T> original, Cloner cloner)
      : base(original, cloner) {
      Content = original.Content; // not sure if to Clone()

    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new Vertex<T>(this, cloner);
    }

    public Vertex(Vertex<T> node) {
      Id = Guid.NewGuid().ToString();
      Label = node.Label;
      Content = node.Content;
      InArcs = new List<IArc>(node.InArcs);
      OutArcs = new List<IArc>(node.OutArcs);
    }
  }
}
